# airflow: gestionnaire de tache de AirBnb

![category](développement)
![tag](python)

Pour gérer des tâches on a l'habitude d'utiliser *cron* pour linux ou le gestionnaire de tache sous windows.

Cela est bien pour lancer une tâche mais pour lancer une suite de tâche et de gérer les erreurs cela
est très compliqué.

La société AirBnb a résolu le problème en créant son propre outil: airflow

## Concept

Airflow se base sur trois concepts:

- DAG & Tasks
- Operators
- Executors

## DAG & Tasks

Un DAG (Directed Acyclic Graph) n’est ni plus ni moins qu’un graphe orienté sans retour possible. 
Nos tâches s’exécuteront donc dans un ordre précis, en parallèle ou à la suite, et ce sans risque de boucle infinie.

![201902091759_1.png](./upload/201902091759_1.png)

Un DAG est lancé de manière régulière ou via un trigger.
Un DAG lancé ou en cours d’exécution est appelé une DagInstance, ses tâches étant donc des TaskInstances.

## Operator

Les tâches d’un DAG permettent de faire de nombreuses choses : lancer un script Bash, vérifier une logique métier dans une fonction Python,
, ou envoyer un mai ou une notification.

La création d’une tâche passe par des operators. A chaque besoin son operator :

- BashOperator : exécute une commande bash,
- PythonOperator : exécute une fonction Python,
- EmailOperator : envoie un mail,
- DockerOperator : exécute une commande dans un container Docker,
- HttpOperator : effectue une requête sur un endpoint HTTP,
- https://github.com/apache/airflow/tree/master/airflow/operators

Il est possible de créer ces propres Operator en créant une classe héritant de BaseOperator avec une fonction execute

## Executor

un executor permet de lancer une tache. IL en existe de nombreux modèle, le modèle par défaut étant SequentialExecutor.
Dans ce modèle les tâches sont lancées les unes derrières les autres en local.

Le modèle LocalExecutor permet de traiter en local des tâches parallèle.

le modèle CeleryExecutor permet le scaling des tâches pourra réellement se faire puisqu’Airflow ira déléguer à Celery la distribution 
des tâches sur un pool de workers. Le scaling sera donc horizontal : plus nous aurons de workers plus 
nous pourrons lancer de tâches en parallèle. Notons tout de même que Celery nécessite l’utilisation d’un 
broker pour l’échange des messages, ainsi qu’une synchronisation du code sur l’ensemble des workers 

## Installation


Encore une fois pour débuter le plus simple est d'utiliser docker et l'image puckel/docker-airflow

On peut retrouver plein d'information d'usage sur ce lien: https://github.com/puckel/docker-airflow

Avant de lancer l'image il est nécessaire de créer un répertoire

- dag: qui va contenir nos DAGs
- plugins: qui va contenir nos plugin si besoin


On peut maintenant lancer une session airflow 

```bash
docker run -d -p 8080:8080 -v $(pwd)/dags:/usr/local/airflow/dags -v $(pwd)/plugins:/usr/local/airflow/plugins --name airflow  puckel/docker-airflow webserver
```

On peut maintenant accéder a airflow via localhost:8080

```note
pour des installations plus compliqué utiliser les docker-compose qu'on trouve sur https://github.com/puckel/docker-airflow
```

Pour ma part j'ai utilisé le docker-compose suivant

```
version: '2.1'
services:
    postgres:
        image: postgres:9.6
        environment:
            - POSTGRES_USER=airflow
            - POSTGRES_PASSWORD=airflow
            - POSTGRES_DB=airflow

    webserver:
        image: puckel/docker-airflow:1.10.2
        restart: always
        depends_on:
            - postgres
        environment:
            - LOAD_EX=n
            - EXECUTOR=Local
        volumes:
            - ./dags:/usr/local/airflow/dags
            # Uncomment to include custom plugins
            #- ./plugins:/usr/local/airflow/plugins
        ports:
            - "8080:8080"
        command: webserver
        healthcheck:
            test: ["CMD-SHELL", "[ -f /usr/local/airflow/airflow-webserver.pid ]"]
            interval: 30s
            timeout: 30s
            retries: 3
```

## Usage

Nous allons écrire notre premier dags.

Il va réaliser le pipeline suivant

.. aafig::
    :textual:

                    +-----+
                    |task1|
                    +--+--+
                       |
            +----------+------+
            |                 |
          +-+---+          +--+--+
          |task2|          |task3|
          +--+--+          +--+--+
             |                |
             |             +--+--+
             |             |task4|
             |             +--+--+
             |                |
             +--------+-------+
                      |
                +-----+---+
                | task5   |
                +---------+

Nous allons écrire  myfirstdag.py dans ./dags

```python
from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from airflow.operators.python_operator import PythonOperator
from datetime import datetime, timedelta
import logging
import random

logger = logging.getLogger(__name__)

default_args = {
    'owner': 'airflow',
    'depends_on_past': False,
    'start_date': datetime(2019, 1, 1),
    'email': ['airflow@example.com'],
    'email_on_failure': False,
    'email_on_retry': False,
    'retries': 1,
    'retry_delay': timedelta(minutes=5),
    # 'queue': 'bash_queue',
    # 'pool': 'backfill',
    # 'priority_weight': 10,
    # 'end_date': datetime(2016, 1, 1),
}

def check_task(ds, **context):
    logger.info("check_task")
    if random.randint(0,1) == 0:
        raise Exception("it's 0")

dag = DAG(
    'my_first_dag', default_args=default_args, schedule_interval=timedelta(minutes=5))

task1 = BashOperator(
    task_id='task1',
    bash_command='date',
    dag=dag)

task2 = BashOperator(
    task_id='task2',
    bash_command='sleep 5',
    dag=dag)

templated_command = """
    {% for i in range(5) %}
        echo "{{ ds }}"
        echo "{{ macros.ds_add(ds, 7)}}"
        echo "{{ params.my_param }}"
    {% endfor %}
"""

task3 = BashOperator(
    task_id='task3',
    bash_command=templated_command,
    params={'my_param': 'Parameter I passed in'},
    dag=dag)

#task4 = PythonOperator(
#    task_id='task4',
#    python_callable=check_task,
#    start_date=datetime(2019, 2, 9)
#    )


task4 = BashOperator(
    task_id='task4',
    bash_command='date',
    dag=dag)


task5 = BashOperator(
    task_id='task5',
    bash_command='date',
    dag=dag)

task1.set_downstream([task2, task3])
task3.set_downstream(task4)
task5.set_upstream([task2, task4])

```

Il est possible de tester de façon unitaire chaque étape

```bash
    docker run exec -it aiflow bash
    airflow test my_first_dag task5 2019-02-09
```

airflow possède en effet un client **airflow** qui permet de faire beaucoup de chose

Sur l'interface visuel vous pouvez lancer, voir, analyser vos dags et leurs réalisations

```warning
j'ai rencontré des problèmes avec PythonOperator sans comprendre vraiment pourquoi
```